#include <windows.h>
#include <stdio.h>

// Uncomment for testing mouse events too.
#define MOUSE_EVENTS_CAPTURED

HANDLE hStdin;
DWORD fdwSaveOldMode;

VOID ErrorExit(LPWSTR);
VOID KeyEventProc(KEY_EVENT_RECORD);
#ifdef MOUSE_EVENTS_CAPTURED
VOID MouseEventProc(MOUSE_EVENT_RECORD);
#endif
VOID ResizeEventProc(WINDOW_BUFFER_SIZE_RECORD);

int main(VOID)
{
    DWORD cNumRead, fdwMode, i;
    INPUT_RECORD irInBuf[128];
    int counter=0;

    // Get the standard input handle.

    hStdin = GetStdHandle(STD_INPUT_HANDLE);
    if (hStdin == INVALID_HANDLE_VALUE)
        ErrorExit(L"GetStdHandle");

    // Save the current input mode, to be restored on exit.

    if (! GetConsoleMode(hStdin, &fdwSaveOldMode) )
        ErrorExit(L"GetConsoleMode");

    wprintf(L"Old Mode 0x%08x\n", fdwSaveOldMode);

    // Enable the window and mouse input events.
    fdwMode = ENABLE_WINDOW_INPUT;
#ifdef MOUSE_EVENTS_CAPTURED
    fdwMode |= ENABLE_MOUSE_INPUT;
#endif
    if (! SetConsoleMode(hStdin, fdwMode) )
        ErrorExit(L"SetConsoleMode");

    // Loop to read and handle the next 100 input events.

    while (counter++ <= 100)
    {
        // Wait for the events.

        if (! ReadConsoleInput(
                hStdin,      // input buffer handle
                irInBuf,     // buffer to read into
                128,         // size of read buffer
                &cNumRead) ) // number of records read
            ErrorExit(L"ReadConsoleInput");

        // Dispatch the events to the appropriate handler.

        for (i = 0; i < cNumRead; i++)
        {
            switch(irInBuf[i].EventType)
            {
                case KEY_EVENT: // keyboard input
                    KeyEventProc(irInBuf[i].Event.KeyEvent);
                    break;

                case MOUSE_EVENT: // mouse input
#ifdef MOUSE_EVENTS_CAPTURED
                    MouseEventProc(irInBuf[i].Event.MouseEvent);
#endif
                    break;

                case WINDOW_BUFFER_SIZE_EVENT: // scrn buf. resizing
                    ResizeEventProc( irInBuf[i].Event.WindowBufferSizeEvent );
                    break;

                case FOCUS_EVENT:  // disregard focus events

                case MENU_EVENT:   // disregard menu events
                    break;

                default:
                    ErrorExit(L"Unknown event type");
                    break;
            }
        }
    }

    // Restore input mode on exit.

    SetConsoleMode(hStdin, fdwSaveOldMode);

    return 0;
}

VOID ErrorExit (LPWSTR lpszMessage)
{
    wprintf(L"%s\n", lpszMessage);

    // Restore input mode on exit.

    SetConsoleMode(hStdin, fdwSaveOldMode);

    ExitProcess(0);
}

VOID KeyEventProc(KEY_EVENT_RECORD ker)
{
    wprintf(L"Key event: ");

    if(ker.bKeyDown)
        wprintf(L"key pressed\n");
    else wprintf(L"key released\n");
}

#ifdef MOUSE_EVENTS_CAPTURED
VOID MouseEventProc(MOUSE_EVENT_RECORD mer)
{
#ifndef MOUSE_HWHEELED
#define MOUSE_HWHEELED 0x0008
#endif
    wprintf(L"Mouse event: ");

    switch (mer.dwEventFlags)
    {
        case 0:
        case DOUBLE_CLICK:
        {
            if (mer.dwButtonState == FROM_LEFT_1ST_BUTTON_PRESSED)
                wprintf(L"left button ");
            else if (mer.dwButtonState == FROM_LEFT_2ND_BUTTON_PRESSED)
                wprintf(L"middle button ");
            else if (mer.dwButtonState == RIGHTMOST_BUTTON_PRESSED)
                wprintf(L"right button ");
            else
                wprintf(L"button ");

            if (mer.dwEventFlags == 0)
            {
                if (mer.dwButtonState == 0)
                    wprintf(L"up\n");
                else
                    wprintf(L"down\n");
            }
            else if (mer.dwEventFlags == DOUBLE_CLICK)
            {
                wprintf(L"double click\n");
            }

            break;
        }
        case MOUSE_MOVED:
            wprintf(L"mouse moved\n");
            break;
        case MOUSE_WHEELED:
            wprintf(L"vertical mouse wheel\n");
            break;
        case MOUSE_HWHEELED:
            wprintf(L"horizontal mouse wheel\n");
            break;
        default:
            wprintf(L"unknown\n");
            break;
    }
}
#endif

VOID ResizeEventProc(WINDOW_BUFFER_SIZE_RECORD wbsr)
{
    wprintf(L"Resize event\n");
    wprintf(L"Console screen buffer is %d columns by %d rows.\n", wbsr.dwSize.X, wbsr.dwSize.Y);
}
